from .bo import BayesOpt
from .ga import GA


def get_range_BO(
    f: callable,
    dimension: int,
    xc_bounds: dict,
    acquisition_function="UCB",
    verbose=False,
    **kwargs,
):
    """compute the range of a black-box function using BayesOpt

    args:
        f: callable function to be optimized

        dimension (int): the number of dimensions of the input space

        xc_bounds (list): each tuple contains the lower and upper bounds for each dimension

        acquisition_function (str or callable, optional): the acquisition function to be used, e.g. 'UCB', 'EI', 'PI'. If None, defaults to 'UCB'.

        verbose (Boolean): if True, prints the optimization progress

        **kwargs: additional keyword arguments for the BayesOpt class

    tip:
        for a less verbose output, use (convergence_curve=False, progress_bar=False)

    return:
        Tuple[Interval, dict]: A tuple containing:
            - response_itvl: The interval of the minimum and maximum from the optimization of the black-box function.
            - opt_hint: A dictionary with optimal input points for the minimum and maximum values.

    """
    from ..pba.intervals.number import Interval

    if not verbose:  # quiet mode
        V = 0

    min_task = BayesOpt(
        f=f,
        dimension=dimension,
        task="minimisation",
        xc_bounds=xc_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )
    min_task.run(verbose=V)
    min_target = min_task.optimal_target

    max_task = BayesOpt(
        f=f,
        dimension=dimension,  # type: ignore
        task="maximisation",
        xc_bounds=xc_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )

    max_task.run(verbose=V)
    max_target = max_task.optimal_target

    # return 1: the interval of min and max
    response_itvl = Interval(min_target, max_target)

    # return 2: the mapping associated with the optimisation
    opt_hint = {
        "min": min_task.optimal,
        "max": max_task.optimal,
    }
    return response_itvl, opt_hint


def get_range_GA(
    f: callable, dimension: int, varbound, algorithm_param=None, verbose=False, **kwargs
):
    """compute the range of the black-box function using GA

    return:
        Tuple[Interval, dict]: A tuple containing:
            - response_itvl: The interval of the minimum and maximum from the optimization of the black-box function.
            - opt_hint: A dictionary with optimal input points for the minimum and maximum values.
    """

    from ..pba.intervals.number import Interval

    if not verbose:
        kwargs["convergence_curve"] = False
        kwargs["progress_bar"] = False

    min_task = GA(f, task="minimisation", dimension=dimension, varbound=varbound)
    min_task.run(algorithm_param=algorithm_param, **kwargs)
    min_target = min_task.optimal_target

    max_task = GA(f, task="maximisation", dimension=dimension, varbound=varbound)
    max_task.run(algorithm_param=algorithm_param, **kwargs)
    max_target = max_task.optimal_target

    # return 1: the interval of min and max
    response_itvl = Interval(min_target, max_target)

    # return 2: the mapping associated with the optimisation
    opt_hint = {
        "min": min_task.optimal,
        "max": max_task.optimal,
    }

    return response_itvl, opt_hint



import numpy as np
from rich.progress import track
from typing import Callable
from scipy.stats import qmc  # Import Latin Hypercube Sampling from SciPy
from pyuncertainnumber.propagation.utils import Propagation_results
from pyuncertainnumber import UncertainNumber
from ...pba.distributions import Distribution


def sampling_aleatory_method(
    xs: list[Distribution],
    f: Callable,
    method: str = "monte_carlo",
    n_sam: int = 1_000,
    results: Propagation_results = None,
    save_raw_data=False,
) -> Propagation_results:  # Specify return type
    #! leslie: I prefer returning the raw output sample
    """Performs aleatory uncertainty propagation using Monte Carlo or Latin Hypercube sampling,
        when its inputs are uncertain and described by probability distributions.

    args:
        xs (list): A list of `UncertainNumber` or `Distribution` objects;
        f (Callable): A callable function that takes input
                  values and returns the corresponding output(s). Can be None,
                  in which case only samples are generated.
        results (Propagation_results, optional): An object to store propagation results.
                                            Defaults to None, in which case a new
                                            `Propagation_results` object is created.
        method (str, optional): The sampling method to use. Choose from:
                            - 'monte_carlo': Monte Carlo sampling (random sampling
                                              from the distributions specified in
                                              the UncertainNumber objects)
                            - 'latin_hypercube': Latin Hypercube sampling (stratified
                                                  sampling for better space coverage)
                            Defaults to 'monte_carlo'.
        n_sam (int): The number of samples to generate for the chosen sampling method.
                Defaults to 1000.
        save_raw_data (str, optional): Acts as a switch to enable or disable the storage of raw
          input data when a function (f) is not provided.
          - 'no': Returns an error that no function is provided.
          - 'yes': Returns the full arrays of unique input combinations.

    signature:
        sampling_aleatory_method(x: list, f: Callable, ...) -> propagation_results

    note:
        If the `f` function returns multiple outputs, the code can accomodate.
        This fuction allows only for the propagation of independent variables.

    returns:
        Returns response samples or a `Propagation_results` object:

        `Propagation_results` object(s) containing:
            - 'un': UncertainNumber object(s) to represent the empirical distribution(s) of the output(s).
            - 'raw_data' (dict): Dictionary containing raw data shared across output(s):
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.

    raises:
        ValueError if no function is provided and save_raw_data is 'no' and if invald UP method is selected.

    example:
        # Example usage with different parameters for minimization and maximization

        >>> results = sampling_aleatory_method(x=x, f=f, n_sam = 300, method = 'monte_carlo', save_raw_data = "no")
    """
    if results is None:
        results = Propagation_results()

    assert callable(f), "The function f must be a callable."

    print(f"Total number of input combinations for the {method} method: {n_sam}")

    match method:
        case "monte_carlo":
            if isinstance(xs[0], Distribution):
                samples = [d.sample(n_sam) for d in xs]
            elif isinstance(xs[0], UncertainNumber):
                samples = [un.construct.sample(n_sam) for un in xs]
            else:
                raise ValueError(
                    "The input must be a list of UncertainNumber or Distribution objects."
                )
            # samples = np.stack(samples, axis=1)
            # samples = np.array([un.random(size=n_sam) for un in x])
        case "latin_hypercube":
            sampler = qmc.LatinHypercube(d=len(xs))
            lhs_samples = sampler.random(n=n_sam)  # u-space (n, d)

            samples = np.concatenate(
                [d.alpha_cut(lhs_samples[:, i]) for i, d in enumerate(xs)], axis=1
            )
        case _:
            raise ValueError("Invalid UP method!")

    # TODO discuss the logic to save verbose raw data @ioannaioan
    # TODO seems to be different behaviors between aleatory and epistemic methods

    if save_raw_data:
        # Transpose to have each row as a sample
        samples = samples.T

        if f is not None:  # Only evaluate if f is provided
            all_output = np.array(
                [
                    f(xi)
                    for xi in track(
                        samples, description="function evaluations for samples"
                    )
                ]
            )

            if all_output.ndim == 1:  # If f returns a single output
                # Reshape to a column vector
                all_output = all_output.reshape(-1, 1)

            # if save_raw_data == 'yes':
            results.add_raw_data(x=samples)
            results.add_raw_data(f=all_output)
        else:
            raise ValueError(
                "No function is provided. Select save_raw_data = 'yes' to save the input combinations."
            )

        return results
    else:
        # simpler logic
        return f(*samples)



import enum

class Measurand(enum.Enum):
    count = 'count'
    tally = 'tally'
    unobservable_parameter = 'unobservable parameter'
    probability = 'probability'
    distribution = 'distribution'
    range_ = 'range'
    rank = 'rank'

''' hint '''
# d = Door('open')
# print(d.value)
# d = Door('closed')
# print(d.value)
# d = Door('is not a valid Door')
# print(d.value)


import enum

class Uncertainty_types(enum.Enum):
    Certain = 'certain'
    Aleatory = 'aleatory'
    Epistemic = 'epistemic'
    Inferential = 'inferential'
    Design_uncertainty = 'design uncertainty'
    Vagueness = 'vagueness'
    Mixture = 'mixture'

''' hint '''
# d = Door('open')
# print(d.value)
# d = Door('closed')
# print(d.value)
# d = Door('is not a valid Door')
# print(d.value)


from .intervals.number import Interval as I
from .pbox_parametric import *
from .pbox_free import *
from .dss import DempsterShafer
from .distributions import Distribution as D
from .distributions import JointDistribution, ECDF
from .dependency import Dependency
from .context import dependency
from .pbox_abc import inspect_pbox



import numpy as np
from typing import Union
import matplotlib.pyplot as plt
from pyuncertainnumber import UncertainNumber


def outward_direction(x: list,
                      n_disc: Union[int, np.ndarray] = 10,
                      tOp: Union[float, np.ndarray] = 0.999,
                      bOt: Union[float, np.ndarray] = 0.001
                      ):  # Specify return type
    """
    args:
        x (list): A list of UncertainNumber objects.
        f (Callable): The function to evaluate.
        results (dict): A dictionary to store the results (optional).
        method (str): The method which will estimate bounds of each combination of focal elements (default is the endpoint)
        lim_Q (np.array): Quantile limits for discretization.
        n_disc (int): Number of discretization points.

    signature:
        second_order_propagation_method(x: list, f: Callable, results: dict, method: str, lim_Q: np.array, n_disc: int) -> dict

    notes:
        Performs second-order uncertainty propagation for mixed uncertain numbers 

    returns:
        dict: A dictionary containing the results
    """
    d = len(x)  # dimension of uncertain numbers
    bounds_x = []
    ranges = np.zeros((2, d))
    u_list = []  # List to store 'u' for each uncertain number

    for i, un in enumerate(x):
        print(f"Processing variable {i + 1} with essence: {un.essence}")

        if un.essence == "distribution":
            # perform outward directed discretisation
            if isinstance(n_disc, int):
                nd = n_disc + 1
            else:
                nd = n_disc[i] + 1  # Use the i-th element of n_disc array

            distribution_family = un.distribution_parameters[0]

            if distribution_family == 'triang':
                u = np.linspace(0, 1, nd)
            else:
                if isinstance(tOp, np.ndarray):
                    top = tOp[i]
                else:
                    top = tOp
                if isinstance(bOt, np.ndarray):
                    bot = bOt[i]
                else:
                    bot = bOt
                u = np.linspace(bot, top, nd)  # Use bOt and tOp here

        elif un.essence == "pbox":
            distribution_family = un.distribution_parameters[0]
            if distribution_family == 'triang':
                if isinstance(n_disc, int):
                    nd = n_disc + 1
                else:
                    nd = n_disc[i] + 1  # Use the i-th element of n_disc array
                u = np.linspace(0, 1, nd)
            else:
                if isinstance(n_disc, int):
                    nd = n_disc + 1
                else:
                    nd = n_disc[i] + 1  # Use the i-th element of n_disc array

                if isinstance(tOp, np.ndarray):
                    top = tOp[i]
                else:
                    top = tOp
                if isinstance(bOt, np.ndarray):
                    bot = bOt[i]
                else:
                    bot = bOt
                u = np.linspace(bot, top, nd)  # Use bOt and tOp here

        else:  # un.essence == "interval"
            u = np.array([0.0, 1.0])  # Or adjust as needed for intervals

        u_list.append(u)  # Add 'u' to the list

        # Generate discrete p-boxes
        match un.essence:
            case "distribution":
                # Calculate xl and xr for distributions (adjust as needed)
                # Assuming un.ppf(u) returns a list or array
                temp_xl = un.ppf(u_list[i][:-1])
                # Adjust based on your distribution
                temp_xr = un.ppf(u_list[i][1:])
                # Create a 2D array of bounds
                rang = np.array([temp_xl, temp_xr]).T
                bounds_x.append(rang)
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])

            case "interval":
                # Repeat lower bound for all quantiles
                temp_xl = np.array([un.bounds[0]])
                # Repeat upper bound for all quantiles
                temp_xr = np.array([un.bounds[1]])
                # Create a 2D array of bounds
                rang = np.array([temp_xl, temp_xr]).T
                bounds_x.append(rang)
                ranges[:, i] = un.bounds  # np.array([un.bounds])

            case "pbox":
                temp_xl = un.ppf(u_list[i][:-1])[0]
                temp_xr = un.ppf(u_list[i][1:])[1]
                # Create a 2D array of bounds
                rang = np.array([temp_xl, temp_xr]).T
                bounds_x.append(rang)
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])

            case _:
                raise ValueError(f"Unsupported uncertainty type: {un.essence}")

    return temp_xl, temp_xr


def plotPbox_pbox(xL, xR,  p=None):
    """
    Plots a p-box (probability box) using matplotlib.

    Args:
        xL (np.ndarray): A 1D NumPy array of lower bounds.
        xR (np.ndarray): A 1D NumPy array of upper bounds.
        p (np.ndarray, optional): A 1D NumPy array of probabilities corresponding to the intervals.
                                   Defaults to None, which generates equally spaced probabilities.
        color (str, optional): The color of the plot. Defaults to 'k' (black).
    """
    xL = np.squeeze(xL)  # Ensure xL is a 1D array
    xR = np.squeeze(xR)  # Ensure xR is a 1D array

    if p is None:
        # p should have one more element than xL/xR
        p = np.linspace(0, 1, len(xL) + 1)

    # Plot the step functions
    plt.step(np.concatenate(([xL[0]], xL)), p, where='post', color='black')
    plt.step(np.concatenate(([xR[0]], xR)), p, where='post', color='red')

    # Add bottom and top lines to close the box
    plt.plot([xL[0], xR[0]], [0, 0], color='red')  # Bottom line
    plt.plot([xL[-1], xR[-1]], [1, 1], color='black')  # Top line

    # # Add the normal distribution
    # mean = 1
    # std = 0.1
    # x_norm = np.linspace(mean - 3.5 * std, mean + 3.5 * std, 100)  # Generate x-values for the normal distribution
    # y_norm = norm.cdf(x_norm, mean, std)  # Calculate the corresponding PDF values
    # plt.plot(x_norm, y_norm, color='blue')  # Plot the normal distribution

    # Add the pbox distribution
    y = UncertainNumber(essence='pbox', distribution_parameters=[
                        "gaussian", [[1.5, 2.5], 0.20]])

    x_l = y._construct.left  # Generate x-values for the normal distribution
    x_r = y._construct.right  # Generate x-values for the normal distribution
    # Calculate the corresponding PDF values
    y_val = np.linspace(0, 1, len(y._construct.left))
    plt.plot(x_l, y_val, color='blue')  # Plot the normal distribution
    plt.plot(x_r, y_val, color='green')  # Plot the normal distribution

    # Add x and y axis labels
    plt.xlabel("U", fontsize=14)
    plt.ylabel("Cumulative Probability", fontsize=14)
    # Increase font size for axis numbers
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)

    # plt.legend()
    # plt.show()
    # # Highlight the points (xL, p)
    # plt.scatter(xL, p, color=colors, marker='o', edgecolors='black', zorder=3)

    # # Highlight the points (xR, p)
    # plt.scatter(xR, p, color=colors, marker='o', edgecolors='black', zorder=3)

    # plt.fill_betweenx(p, xL, xR, color=colors, alpha=0.5)
    # plt.plot( [xL[0], xR[0]], [0, 0],color=colors, linewidth=3)
    # plt.plot([xL[-1], xR[-1]],[1, 1],  color=colors, linewidth=3)
    plt.show()


def plotPbox(xL, xR, p=None):
    """
    Plots a p-box (probability box) using matplotlib.

    Args:
        xL (np.ndarray): A 1D NumPy array of lower bounds.
        xR (np.ndarray): A 1D NumPy array of upper bounds.
        p (np.ndarray, optional): A 1D NumPy array of probabilities corresponding to the intervals.
                                   Defaults to None, which generates equally spaced probabilities.
        color (str, optional): The color of the plot. Defaults to 'k' (black).
    """
    xL = np.squeeze(xL)  # Ensure xL is a 1D array
    xR = np.squeeze(xR)  # Ensure xR is a 1D array

    if p is None:
        # p should have one more element than xL/xR
        p = np.linspace(0, 1, len(xL) + 1)

    # Plot the step functions
    plt.step(np.concatenate(([xL[0]], xL)), p, where='post', color='black')
    plt.step(np.concatenate(([xR[0]], xR)), p, where='post', color='red')

    # Add bottom and top lines to close the box
    plt.plot([xL[0], xR[0]], [0, 0], color='red')  # Bottom line
    plt.plot([xL[-1], xR[-1]], [1, 1], color='black')  # Top line

    # Add x and y axis labels
    plt.xlabel("X", fontsize=14)
    plt.ylabel("Cumulative Probability", fontsize=14)
    # Increase font size for axis numbers
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)

    plt.show()


means = np.array([1, 2, 3, 4, 5])
stds = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

n_disc = 10  # Number of discretizations

x = [
    UncertainNumber(essence='distribution', distribution_parameters=[
                    "gaussian", [means[0], stds[0]]])

    #  UncertainNumber(essence = 'interval', bounds= [means[1]-2* stds[1], means[1]+2* stds[1]]),
    # UncertainNumber(essence = 'interval', bounds= [means[2]-2* stds[2], means[2]+2* stds[2]])
]

xl, xr = outward_direction(x)

# plotPbox(xl, xr, p=None)
# plt.show()


x0 = [
    UncertainNumber(essence='pbox', distribution_parameters=[
                    "gaussian", [[1.5, 2.5], 0.20]])
    # UncertainNumber(essence = 'interval', bounds = [2.5, 3.5])

]

y = UncertainNumber(essence='pbox', distribution_parameters=[
                    "gaussian", [[0.5, 1.5], 0.10]])
print('y', y._construct.left)
xl, xr = outward_direction(x0)
plotPbox_pbox(xl, xr, p=None)
plt.show()


def plot_interval(lower_bound, upper_bound, color='blue', label=None):
    """
    Plots an interval on a Matplotlib plot.

    Args:
        lower_bound (float): The lower bound of the interval.
        upper_bound (float): The upper bound of the interval.
        color (str, optional): The color of the interval line. Defaults to 'blue'.
        label (str, optional): The label for the interval in the legend. Defaults to None.
    """

    plt.plot([lower_bound, upper_bound], [1, 1], color=color,
             linewidth=2, label=label)  # Plot the interval line
    plt.plot([lower_bound, lower_bound], [0, 1], color=color,
             linewidth=2, label=label)  # Plot the interval line
    plt.plot([upper_bound, upper_bound], [0, 1], color=color,
             linewidth=2, label=label)  # Plot the interval line

    # Add x and y axis labels
    plt.xlabel("V", fontsize=14)
    plt.ylabel("Cumulative Probability", fontsize=14)
    # Increase font size for axis numbers
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)


y = UncertainNumber(essence='interval', bounds=[2.5, 3.5])
plot_interval(2.5, 3.5)
plt.show()



from bayes_opt import BayesianOptimization
from bayes_opt import acquisition
import numpy as np
import inspect
from functools import partial
import warnings


class BayesOpt:
    """Bayesian Optimisation class

    args:
        f (callable): the target function to be optimised, should have a single argument

        xc_bounds (dict): the bounds for the design space, e.g. {'x1': (0, 1), 'x2': (0, 1)}

        dimension (int): the dimension of the design space, i.e. the number of parameters

        task (str): either 'minimisation' or 'maximisation'

        acquisition_function (str or callable, optional): the acquisition function to be used, e.g. 'UCB', 'EI', 'PI'. If None, defaults to 'UCB'.

        num_explorations (int, optional): the number of initial exploration points. Defaults to 100.

        num_iterations (int, optional): the number of iterations to run the optimisation. Defaults to 100.

    note:
        Acquisition functions can be either a string (e.g. 'UCB', 'EI', 'PI') or a callable function.
        'UCB' stands for Upper Confidence Bound, 'EI' for Expected Improvement, and 'PI' for Probability of Improvement.
        If a string is provided, the parameter for the acquisition function can be passed as an additional
        argument to the class constructor. For example, for 'UCB', you can pass a `kappa` value, and for 'EI' or 'PI', you can pass an `xi` value.
        For low-level controls, if a callable function is provided, it should already be parameterised.

    example:
        >>> import numpy as np
        >>> from pyuncertainnumber.opt.bo import BayesOpt
        >>> def black_box_function(x):
        ...     return np.exp(-(x - 2)**2) + np.exp(-(x - 6)**2 / 10) + 1 / (x**2 + 1)
        >>> bo = BayesOpt(
        ...     f=black_box_function,
        ...     dimension=1,
        ...     xc_bounds={'x': (-2, 10)},
        ...     task='maximisation',
        ...     num_explorations=3,
        ...     num_iterations=20
        ... )
        >>> bo.run(verbose=True)
        >>> print(bo.optimal)  # get the optimal parameters and target value

    .. admonition:: Implementation

        For 2D (or higher) optimisation task, the variable bounds can be specified using `EpistemicDomain` class.

        example:
            >>> from from pyuncertainnumber import pba, EpistemicDomain
            >>> ed = EpistemicDomain(pba.I(-5, 5), pba.I(-5, 5))
            >>> BayesOpt(f=foo,
            ...     dimension=2,
            ...     xc_bounds= ed.to_BayesOptBounds(),  # the trick
            ...     task='maximisation',
            ...     num_explorations=3,
            ...     num_iterations=20,
            ...     )
    """

    # TODO add a descriptor for `task`
    def __init__(
        self,
        f,
        xc_bounds,
        dimension,
        task,
        acquisition_function="UCB",
        num_explorations=100,
        num_iterations=100,
    ):

        self.task = task  # either minimisation or maximisation
        self.num_explorations = num_explorations  # initial exploration points
        self.num_iterations = num_iterations
        self.xc_bounds = xc_bounds
        self.dimension = dimension
        self.acquisition_function = self.parse_acq(acquisition_function)
        self.f = f  # the function to be optimised

    def parse_acq(self, acq, parameter=None):
        """parse the acquisition function

        args:
            acq (str or callable): the acquisition function to be used. See notes above.
            parameter (float, optional): parameter for the acquisition function, e.g. kappa for UCB, xi for EI and PI
        """
        if isinstance(acq, str):
            if acq == "UCB":
                kappa = parameter if parameter else 10.0
                return acquisition.UpperConfidenceBound(kappa=kappa)
            elif acq == "EI":
                xi = parameter if parameter else 0.01
                return acquisition.ExpectedImprovement(xi=xi)
            elif acq == "PI":
                xi = parameter if parameter else 0.01
                return acquisition.ProbabilityOfImprovement(xi=xi)
            else:
                raise ValueError(f"Unknown acquisition function: {acq}")
        else:
            return acq

    @property
    def f(self):
        """return the function to be optimised"""
        return self._f

    @f.setter
    def f(self, f):
        # step 1: dimension check
        if not check_argument_count(f) == "Single argument":
            warnings.warn(
                "The function to be optimised should have a single argument",
                category=RuntimeWarning,
            )
        if self.dimension > 1 and (check_argument_count(f) == "Single argument"):
            f = partial(transform_func, fb=f)
        # step 2: flip check by the task
        # the first flip is to make the function minimisation
        if self.task == "maximisation":
            self._f = f
        elif self.task == "minimisation":
            self._f = lambda *args, **kwargs: -f(*args, **kwargs)

    def get_results(self):
        """inspect the results, to save or not"""

        self._optimal_dict = {}
        # TODO serialise the dict, plus the undering GP model

        if self.task == "maximisation":
            bo_all_dict = {
                "Xc_params": self.optimizer.space.params.tolist(),
                "target_array": self.optimizer.space.target.tolist(),
                "optimal_Xc": list(self.optimizer.max["params"].values()),
                "optimal_target": self.optimizer.max["target"],
            }
        elif self.task == "minimisation":  # the second flip
            target_arr = self.optimizer.space.target.copy()
            target_arr[:] *= -1

            optimal_index = np.argmin(target_arr)
            optimal_target = np.min(target_arr)
            optimal_Xc = self.optimizer.space.params[optimal_index]

            bo_all_dict = {
                "Xc_params": self.optimizer.space.params.tolist(),
                "target_array": target_arr.tolist(),
                "optimal_Xc": optimal_Xc.tolist(),
                "optimal_target": optimal_target.tolist(),
            }

        self._optimal_dict["xc"] = bo_all_dict["optimal_Xc"]
        self._optimal_dict["target"] = bo_all_dict["optimal_target"]
        self._all_results = bo_all_dict

    def run(self, **kwargs):
        """run the Bayesian optimisation process.

        args:
            verbose (bool, optional): whether to print the progress. Defaults to False. Use 'verbose=True' to see the progress.

            **kwargs: additional low--level arguments to be passed to the BayesianOptimization constructor.

        example:
            >>> foo.run(verbose=True)
        """

        self.optimizer = BayesianOptimization(
            f=self.f,
            pbounds=self.xc_bounds,
            acquisition_function=self.acquisition_function,
            random_state=42,
            allow_duplicate_points=True,
            **kwargs,
        )

        try:
            # initial exploration of the design space
            self.optimizer.maximize(
                init_points=self.num_explorations,
                n_iter=0,
            )
        except:
            pass

        # * _________________ run the BO iterations to get the optimal Xc
        for _ in range(self.num_iterations):
            next_point = self.optimizer.suggest()
            target = self._f(**next_point)
            self.optimizer.register(params=next_point, target=target)
            # print(target, next_point)

        # * _________________ compile the results
        self.get_results()

    @property
    def optimal(self) -> dict:
        """return the optimal parameters and target value as a dictionary"""
        return self._optimal_dict

    @property
    def optimal_xc(self) -> np.ndarray:
        """return the optimal design points (xc)"""
        return np.squeeze(self._optimal_dict["xc"])

    @property
    def optimal_target(self) -> np.ndarray:
        """return the optimal target value f(xc*)"""
        return self._optimal_dict["target"]


def check_argument_count(func):
    # Get the function signature
    sig = inspect.signature(func)
    # Count the number of non-default parameters
    param_count = len(
        [p for p in sig.parameters.values() if p.default == inspect.Parameter.empty]
    )
    return "Single argument" if param_count == 1 else "Multiple arguments"


def transform_func(fb, **kwargs):
    args = [p for p in kwargs.values()]
    return fb(args)



import enum

class Variability(enum.Enum):
    point_estimate = 'point estimate'
    confidence = 'Confidence'


''' hint '''
# d = Door('open')
# print(d.value)
# d = Door('closed')
# print(d.value)
# d = Door('is not a valid Door')
# print(d.value)


import enum


class Ensemble(enum.Enum):
    repeated_measurements = "repeated measurements"
    flights = "flights"
    pressurisations = "pressurisations"
    temporal_steps = "temporal steps"
    spatial_sites = "spatial sites"
    manufactured_components = "manufactured components"
    customers = "customers"
    people = "people"
    households = "households"
    particular_population = "particular population"

    # TODO user can edit at will...


""" hint """
# d = Door('open')
# print(d.value)
# d = Door('closed')
# print(d.value)
# d = Door('is not a valid Door')
# print(d.value)
